/*
 * $Id: safebuf.c,v 1.5 2005/05/01 12:36:41 oldgit Exp $
 * safe character buffers for RTF-to-LaTeX2e translation writer code.
 * (c) 2005 Richard Bornat
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include "rtf.h" // for RTFPanic
#include "safebuf.h" // wherein see comment about how this all came about

#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define MIN(a,b) ((a)<(b) ? (a) : (b))

SafeBuf safebuf_new(size_t size) {
    SafeBuf buf;
    if ((buf=(SafeBuf)malloc(sizeof(*buf)))==NULL)
	RTFPanic("safebuf_new can't alloc buf");
    if ((buf->_buf=malloc(size))==NULL)
	RTFPanic("safebuf_new can't alloc buf->_buf %d", size);
    buf->_idx=0; buf->_size=size;
    buf->_buf[0]='\0';
    return buf;
}

void safebuf_dispose(SafeBuf buf) {
    free(buf->_buf);
    free(buf);
}

void safebuf_clear(SafeBuf buf) {
    buf->_idx = 0; buf->_buf[0]='\0';
}

void safebuf_expand(SafeBuf buf, size_t n) {
    char *_buf;
    size_t _size = buf->_size+MAX(n, MIN(buf->_size, 1024));
    if ((_buf=malloc(_size))==NULL)
	RTFPanic("buf_expand failed (%d)", _size);
    strncpy(_buf, buf->_buf, buf->_idx);
    free(buf->_buf);
    buf->_buf = _buf;
    buf->_size = _size;
}

void safebuf_append_ch(SafeBuf buf, char c) {
    if (buf->_idx==buf->_size)
	safebuf_expand(buf, 1);
    buf->_buf[buf->_idx++] = c;
}

void safebuf_append_str(SafeBuf buf, char *str) {
    size_t n = strlen(str);
    if (buf->_idx+n+1>buf->_size)
	safebuf_expand(buf, buf->_idx+n+1-buf->_size);
    strncpy(buf->_buf+buf->_idx, str, n);
    buf->_idx += n;
    buf->_buf[buf->_idx] = '\0'; // cos that's what strcat does
}

void safebuf_assign_str(SafeBuf buf, char *str) {
    buf->_idx = 0;
    safebuf_append_str(buf, str);
}

int safebuf_ncmp(SafeBuf buf, char *s, int n) {
    if (n>buf->_size)
	RTFPanic("safebuf_ncmp %d %d", buf->_size, n);
    return strncmp(buf->_buf, s, n);
}

char *safebuf_contents(SafeBuf buf) {
    char *res;
    buf->_buf[buf->_idx] = '\0';
    if ((res=RTFStrSave(buf->_buf))==NULL)
	RTFPanic("cannot save \"%s\"", buf->_buf);
    return res;
}

int safebuf_sprintf(SafeBuf buf, char *fmt, ...) {
    do {
	int avail = buf->_size-buf->_idx, printed;
	va_list args;
	if (avail==0) {
	    safebuf_expand(buf, 1); avail=1;
	}
	va_start(args, fmt);
	printed = vsnprintf(buf->_buf+buf->_idx, avail, fmt, args);
	va_end(args);
	if (printed+1>avail)
	    safebuf_expand(buf, printed+1-avail);
	else {
	    buf->_idx += printed;
	    return printed;
	}
    } while(1);
    
    RTFPanic("exit from safebuf_sprintf loop"); // can't happen
    return  -1; // neither can that
} 